#pragma implementation

#include "Instantiator.h"
#include "JavaClassFile.h"
#include "JavaFieldSignature.h"
#include "JavaMethodSignature.h"
#include "PJavaConstraint.h"
#include "AuxMethodSignature.h"
#include "AuxMethodMap.h"
#include "find_leaks.h"
#include <iostream.h>

//
//  Method name : (Constructor)
//  Description : Create an instantiator that replaces all parameters named
//      pName with "type."
//
CInstantiator::CInstantiator(const unicode_string& pName,
			     const CJavaTypeSignature& type)
{
  REF_PTR(this);
  fMap.insert(InstantiationMap::value_type(pName, type));
}

//
//  Method name : (Constructor)
//  Description : Create an instantiator to instantiate type "ptype" on
//      params.  
//  Checks: params.size() is equal to ptype.fPClassInfo->fParams.size(). 
//         "ptype" need not be parameterized.
//
CInstantiator::CInstantiator(const CJavaClassFile& ptype,
			     const deque<CJavaTypeSignature>* params = 0)
{
  REF_PTR(this);
  if (params == 0 || params->size() == 0) {
    if (ptype.IsParameterized()) {
      // This should never happen
      cerr << "***Bad instantiation***:"
	   << ptype.GetClassName()
	   << "[ ]" << endl
	   << "You have found a bug in the compiler." << endl << endl
	   << "(Somewhere within your code, you used the parameterized class"
	   << ptype.GetClassName() << endl
	   << "without any parameters.  Ordinarily, PolyJ would detect this, "
	   << endl 
	   << "but it seems that you've found a case where it doesn't.  Please"
	   << endl
	   << "send a bug report to <polyj-bugs@www.pmg.lcs.mit.edu>."
	   << endl;      
      assert(0);
    }
  } else {
    assert(ptype.GetPInfo().fParams);
    deque<unicode_string>::const_iterator name_iter = 
      ptype.GetPInfo().fParams->begin();
    deque<unicode_string>::const_iterator name_end =
      ptype.GetPInfo().fParams->end();
    deque<CJavaTypeSignature>::const_iterator type_iter = params->begin();
    if (params->size() != ptype.GetPInfo().fParams->size()) {
      // This should never happen either.
      cerr << "***Bad instantiation***:" << endl
	   << "    "
	   << ptype.GetClassName()
	   << "[";
      for( ; type_iter != params->end(); ) {
	cerr << (*type_iter);
	if (++type_iter != params->end()) cerr << ",";
      }
      cerr << "]" << endl << "for " << ptype.GetType() << endl
	   << "You have found a bug in the compiler." << endl << endl
	   << "(Somewhere within your code, you used the parameterized class"
	   << ptype.GetClassName() << endl
	   << "without the wrong number of parameters.  Ordinarily, PolyJ"
	   << endl 
	   << "would detect this, but it seems that you've found a case where"
	   << endl
	   << "it doesn't.  Please send a bug report to <polyj-bugs@www.pmg.lcs.mit.edu>."
	   << endl;      

      assert(0);
    }
    for( ; name_iter != name_end; ++name_iter, ++type_iter) {
      fMap.insert(InstantiationMap::value_type(*name_iter,*type_iter));
    }
  }
}

//
// (Copy constructor)
//
CInstantiator::CInstantiator(const CInstantiator &other)
  : fMap(other.fMap)
{
  REF_PTR(this);
}

//
// (Destructor)
//
CInstantiator::~CInstantiator()
{
  UNREF_PTR(this);
}

//
//  Method name: Instantiate
//  Description: Sets "fill" to the type signature for the parameter named by
//      "param". 
//  Requires: There is a mapping for "Param".
void
CInstantiator::Instantiate(CJavaTypeSignature& fill,
			   const unicode_string& param) const
{
  InstantiationMap::const_iterator found =
    fMap.find(param);
  if (found == fMap.end()) {
    cerr << "Dying on " << ::UnicodeToUTF(param) << ":" << endl;
    for (InstantiationMap::const_iterator i = fMap.begin();
	 i != fMap.end(); ++i) {
      cerr << "  " << ::UnicodeToUTF((*i).first) << " -> "
	   << (*i).second << endl;
    }
    assert(0);
  } else {
    fill = (*found).second;
  }
}

//
//  Method name: Instantiate
//  Description: Set fill to the type signature for the instantiation of "type"
//
void
CInstantiator::Instantiate(CJavaTypeSignature& fill,
			   const CJavaTypeSignature& type) const
{
  if (type.GetBaseType() == CJavaTypeSignature::Parameter) {
    unsigned long bounds = type.fArrayBounds;
    Instantiate(fill, type.fClassName); 
    fill.fArrayBounds += bounds;
  } else {
    if (&fill != &type)
      fill = type;
    if (type.IsParameterized()) {
      for(deque<CJavaTypeSignature>::iterator i = fill.fParameters->begin();
	  i != fill.fParameters->end(); ++i) {
	Instantiate(*i, *i);
      }
    }
  }
}  

//
//  Method name: Instantiate
//  Description: Set fill to the signature for the instantiation of field.
//
void 
CInstantiator::Instantiate(CJavaFieldSignature& fill, 
			   const CJavaFieldSignature& field) const
{
  fill = field;
  Instantiate(fill.fType, field.fType);
  fill.SetHashKey();
}

//
//  Method name: Instantiate
//  Description: Set fill to the signature for the instantiation of method.
//
void
CInstantiator::Instantiate(CJavaMethodSignature& fill,
			   const CJavaMethodSignature& method) const
{ 
  Instantiate(fill.fReturnType, method.fReturnType);
  if (&fill != &method) {
    fill.fMethodName = method.fMethodName;
    fill.fArgumentTypes = method.fArgumentTypes;
  }
  for(deque<CJavaTypeSignature>::iterator i = fill.fArgumentTypes.begin();
      i != fill.fArgumentTypes.end(); ++i) {
    Instantiate(*i, *i);
  }
  fill.SetHashKey();
}

//
//  Method name: Instantiate
//  Description: Set fill to be the instantation of "constraint".
//  Requires: &fill != &constraint.
//
void 
CInstantiator::Instantiate(CPJavaConstraint& fill, 
			   const CPJavaConstraint& constraint) const
{
  assert(&fill != &constraint);
  // Remove old parameters
  for (deque<CPJavaConstraint::ConstraintMethod*>::iterator i 
	 = fill.fMethods.begin(); i != fill.fMethods.end(); ++i) 
    delete *i;
  fill.fMethods.erase(fill.fMethods.begin(),fill.fMethods.end());
  fill.fParam = constraint.fParam;
  for(deque<CPJavaConstraint::ConstraintMethod*>::const_iterator m = 
	constraint.fMethods.begin(); m != constraint.fMethods.end(); ++m) {
    CPJavaConstraint::ConstraintMethod* instantiated
      = new CPJavaConstraint::ConstraintMethod;
    instantiated->fStatic = (*m)->fStatic;
    if ((*m)->fThrows) {
      instantiated->fThrows = new deque<CJavaTypeSignature>;
      for(deque<CJavaTypeSignature>::iterator i = (*m)->fThrows->begin();
	  i != (*m)->fThrows->end(); ++i) {
	CJavaTypeSignature iType;
	Instantiate(iType, (*i));
	instantiated->fThrows->push_back(iType);
      }
    } else {
      instantiated->fThrows = new deque<CJavaTypeSignature>;
    }
    instantiated->fMethod = new CJavaMethodSignature;
    Instantiate(* instantiated->fMethod,* (*m)->fMethod);
    fill.fMethods.push_back(instantiated);
  }
}

//
//  Method name: Instantiate
//  Description: Set fill to the instantation of auxMethod
//
void 
CInstantiator::Instantiate(CAuxMethodSignature& fill,
			   const CAuxMethodSignature& auxMethod) const
{
  assert(auxMethod.repOK());
  fill.fType = auxMethod.fType;
  fill.fWType = auxMethod.fWType;
  fill.fWidth = auxMethod.fWidth;
  fill.fDim = auxMethod.fDim;

  Instantiate(fill.fClass, auxMethod.fClass);

  if (&fill != &auxMethod) {
    delete fill.fMethod;
    delete fill.fFirstMethod;
    delete fill.fThrows;
    delete fill.fName;
    delete fill.fFirstClass;
    if (auxMethod.fName != 0)
      fill.fName = new unicode_string(*auxMethod.fName);
    else
      fill.fName = 0;
    if (auxMethod.fFirstClass != 0)
      fill.fFirstClass = new CJavaTypeSignature(*auxMethod.fFirstClass);
    else
      fill.fFirstClass = 0;
    if (auxMethod.fFirstMethod != 0)
      fill.fFirstMethod = new CJavaMethodSignature(*auxMethod.fFirstMethod);
    else
      fill.fFirstMethod = 0;
  }
  
  if (auxMethod.fMethod != 0) {
    assert(auxMethod.fFirstMethod && auxMethod.fThrows);
    if (&fill != &auxMethod) {
      fill.fMethod = new CJavaMethodSignature;
      fill.fThrows = new deque<CJavaTypeSignature>(*auxMethod.fThrows);
    }
    Instantiate(*fill.fMethod,*auxMethod.fMethod);
  } else {
    fill.fMethod = 0;
    fill.fThrows = 0;
  }
     
  assert(fill.repOK());
}

//
//  Method name: Instantiate
//  Description: Add the instantiated auxmethods from map to fill. Do _not_
//      remove existing methods in fill.
//
void 
CInstantiator::Instantiate(CAuxMethodMap& fill, const CAuxMethodMap& map) const
{
  for(amm_rep::const_iterator i = map.Begin();
      i != map.End(); ++i) {
    amm_rep::value_type v;
    // Discard "const".
    Instantiate((CAuxMethodSignature &) v.first,(*i).first);
    unicode_string *str = new unicode_string(*(*i).second);
    v.second = str;
    fill.fMap.insert(v);
  }
}

//
// Method name : Compare
// Description : Returns <0 if the instantiation of param is less than ref,
//    0 if equal, and >0 if greater.  Faster than 
//  {CJavaTypeSignature foo; Instantiate(foo,param); return foo.Compare(ref);} 
//
int 
CInstantiator::Compare(const CJavaTypeSignature& param, 
		       const CJavaTypeSignature& ref) const
{
  if (param.GetBaseType() == CJavaTypeSignature::Parameter) {
    InstantiationMap::const_iterator found =
      fMap.find(param.fClassName);
    return (*found).second.Compare(ref);
  }
  int result = param.GetBaseType() - ref.GetBaseType(); 
  if (result == 0) {
    result = param.GetArrayBounds() - ref.GetArrayBounds();
    if (result == 0 && param.GetBaseType() == CJavaTypeSignature::Class) {
      result = param.fClassName.compare(ref.fClassName);
      if (result == 0 && param.IsParameterized()) {
	if (! ref.IsParameterized()) { return -1; }
	result = ref.GetNParameters() - param.GetNParameters();
	if (result == 0) {
	  deque<CJavaTypeSignature>::const_iterator i = 
	    param.GetParameters()->begin();
	  deque<CJavaTypeSignature>::const_iterator j = 
	    ref.GetParameters()->begin();
	  while (result == 0 && i != param.GetParameters()->end()) {
	    result = Compare(*i,*j);
	    ++i; ++j;
	  }
	}
      }
    }
  }
  return result; 
}

//
//  Method name : Compare
//  Description : Returns 0 if param, when parameterized, is identical to
//    ref, returns -1 if param is greater, and 1 if param is less than ref. 
//    Faster than
//    {CJavaFieldSignature f; Instantiate(f,ref); return f.Compare(ref); }
//
int 
CInstantiator::Compare(const CJavaFieldSignature& param, 
		       const CJavaFieldSignature& ref) const
{
  int result = param.fFieldName.compare(ref.fFieldName);
  if (result == 0) {
    return Compare(param.fType, ref.fType);
  }
  return result;
}

//
//  Method name : Compare
//  Description : Compares the arguments, names, and return types of two 
//    CJavaMethodSignature objects, instantiating the first, and returns 0 if 
//    they are equal, -1 if param is greater than ref and 1 if ref is greater 
//    than param.  Faster than
//    {CJavaMethodSignature m; Instantiate(m,param); return m.Compare(ref);} 
//
int 
CInstantiator::Compare(const CJavaMethodSignature& param, 
		       const CJavaMethodSignature& ref) const
{
  int result = 0;
  if (result == 0) {
    result = param.fMethodName.compare(ref.fMethodName);
    if (result == 0) {
      result = Compare(param.fReturnType, ref.fReturnType) ||
	(ref.fArgumentTypes.size() - param.fArgumentTypes.size());
      if (result == 0) {
	deque<CJavaTypeSignature>::const_iterator i1 = 
	  param.fArgumentTypes.begin();
	deque<CJavaTypeSignature>::const_iterator i2 =
	  ref.fArgumentTypes.begin();
	for ( ; i1 != param.fArgumentTypes.end(); ++i1, ++i2) {
	  if ((result = Compare(*i1, *i2)) != 0)
	    break;
	}
      }
    }
  }
  return result;
}

//
//  Method name : EqualsIgnoreReturn
//  Description : Returns true if the name and arguments of param, when
//    instantiated, match the name and arguments of ref.  The return types 
//    do not affect the result of this function.   Faster than
//      { CJavaMethodSignature m; Instantiate(m,param); 
//        return m.EqualsIgnoreReturn(ref); } 
//
bool 
CInstantiator::EqualsIgnoreReturn(const CJavaMethodSignature& param, 
				  const CJavaMethodSignature& ref) const
{
  bool equal = param.fMethodName == ref.fMethodName &&
    param.fArgumentTypes.size() == ref.fArgumentTypes.size();
  if (equal) {
    deque<CJavaTypeSignature>::const_iterator i1 = 
      param.fArgumentTypes.begin();
    deque<CJavaTypeSignature>::const_iterator i2 =
      ref.fArgumentTypes.begin();
    for (; equal && i1 != param.fArgumentTypes.end(); ++i1, ++i2) {
      equal = Equals(*i1,*i2);
    }
  }
  return equal;
}

//
// Method name : Disassemble
// Description : Writes a representation of this object to os.
//
void 
CInstantiator::Disassemble(ostream &os) const
{
  os << "{";
  for (InstantiationMap::const_iterator i = fMap.begin();
       i != fMap.end(); ++i) {
    os << "(" << ::UnicodeToUTF((*i).first) << "=>"
       << ((*i).second) << ")";
  }
  os << "}";
}
