// Copyright (c) 1996  David Engberg  All rights reserved
#pragma implementation
#include "find_leaks.h"
#include "JavaClassFile.h"
#include "JavaDirectory.h"
#include "config.h"
#include <fstream.h>

#ifdef HAS_IOS_BINARY
#define IOS_BINARY ios::binary
#else
#define IOS_BINARY ios::bin
#endif

extern "C" {
  #include <sys/types.h>
  #include <sys/stat.h>
  #include <fcntl.h>
  #include <unistd.h>
  #include <stdio.h>
}
#include "mapped_file.h"

//
//  Method name : CJavaDirectory
//  Description : Default constructor.  Required for list<CJavaDirectory>.
//    Don't call directly.
//
CJavaDirectory::CJavaDirectory(bool parameterized)
  : fBasePath(""),
    fZipFile(0),
    fIsValid(false),
    fParameterized(parameterized)
{
}

//
//  Method name : CJavaDirectory
//  Description : Constructs a java directory corresponding to the provided
//    base directory name.  This can either be a real directory on disk or
//    a path to a zip file.
//
CJavaDirectory::CJavaDirectory(const string& baseDirectory, bool parameterized)
  : fBasePath(baseDirectory),
    fZipFile(0),
    fIsValid(false),
    fParameterized(parameterized),
    fZipDate(0)
{
  if (fBasePath.IsReadable() && fBasePath.IsDirectory()) {
    fIsValid = true;
  } else if (fBasePath.IsFile() && baseDirectory.length() > 4 &&
	     (baseDirectory.compare(".jar", baseDirectory.length() - 4) == 0
	   || baseDirectory.compare(".zip", baseDirectory.length() - 4) == 0)) {
    fZipFile = new ZipFile;
    fZipFile->reference_count = 1;
    FILE* infile = fopen(baseDirectory.c_str(), "rb");
    fZipFile->fd = fileno(infile);
    if (fZipFile->fd >= 0 && ::read_zip_archive(fZipFile) == 0) {      
      fIsValid = true;
      struct stat st;
      fstat(fZipFile->fd, &st);
      fZipDate = st.st_mtime;
    } else {
      delete fZipFile;
      fZipFile = 0;
    }
  }
}

//
//  Method name : CJavaDirectory
//  Description : Copy constructor.
//
CJavaDirectory::CJavaDirectory(const CJavaDirectory& source)
  : fBasePath(source.fBasePath),
    fRelativePath(source.fRelativePath),
    fZipFile(source.fZipFile),
    fIsValid(source.fIsValid),
    fParameterized(source.fParameterized)
{
  if (fZipFile != 0) {
    fZipFile->reference_count++;
  }
}

//
//  Method name : CJavaDirectory
//  Description : Creates a relative java directory, indicating some package
//    directory that is relative to the provided base.
//
CJavaDirectory::CJavaDirectory(const CJavaDirectory& base,
			       const string& relativePath)
  : fBasePath(base.fBasePath),
    fRelativePath(base.fRelativePath + relativePath + "/"),
    fZipFile(base.fZipFile),
    fIsValid(base.fIsValid),
    fParameterized(base.fParameterized)
{
  if (fZipFile != 0) {
    fZipFile->reference_count++;
  }
  if (fIsValid) {
    if (fZipFile == 0) {
      CFilePath fullPath(fBasePath.GetString() + "/" + fRelativePath);
      fIsValid = fullPath.IsDirectory() && fullPath.IsReadable();
    } else {
      fIsValid = false;
      ZipDirectory* directory = (ZipDirectory*)fZipFile->central_directory;
      for (int i = 0; i < fZipFile->count;
	   ++i, directory = ZIPDIR_NEXT(directory)) {
	const char* zipFile = ZIPDIR_FILENAME(directory);
	if (fRelativePath.compare(zipFile, 0, fRelativePath.length()) == 0) {
	  fIsValid = true;
	  break;
	}
      }
    }
  }
}

//
//  Method name : ~CJavaDirectory
//  Description : Destructor
//
CJavaDirectory::~CJavaDirectory()
{
  if (fZipFile != 0 && --fZipFile->reference_count == 0) {
    ::close(fZipFile->fd);
    free(fZipFile->central_directory);
    delete fZipFile;
  }
}

//
//  Method name : operator=
//  Description : Assignment operator.
//
CJavaDirectory&
CJavaDirectory::operator=(const CJavaDirectory& source)
{
  if (&source != this) {
    fBasePath = source.fBasePath;
    fRelativePath = source.fRelativePath;
    fZipFile = source.fZipFile;
    fIsValid = source.fIsValid;
    if (fZipFile != 0) {
      fZipFile->reference_count++;
    }
    fParameterized = source.fParameterized;
  }
  return *this;
}

//
//  Method name : operator==
//  Description : Compares this directory against another, returns true if they
//    are identical.
//
bool
CJavaDirectory::operator==(const CJavaDirectory& other) const
{
  return fBasePath == other.fBasePath &&
    fRelativePath == other.fRelativePath &&
    fZipFile == other.fZipFile;
}

//
//  Method name : operator<
//  Description : This method is required to put CJavaDirectory in a list<>.
//    It doesn't make much sense by itself.
//
bool
CJavaDirectory::operator<(const CJavaDirectory& other) const
{
  return fBasePath.GetString() < other.fBasePath.GetString() ||
    fRelativePath < other.fRelativePath;
}

//
//  Method name : LoadClassFile
//  Description : This method attempts to load the java class file with the
//    provided name, relative to this java directory.  If this operation could
//    be completed, this returns the newly constructed class file object,
//    otherwise it returns 0.
//
CJavaClassFile*
CJavaDirectory::LoadClassFile(const unicode_string& className) const
{
  CJavaClassFile* result = 0;
  if (IsValid()) {
    string classFileName =
      fRelativePath + ::UnicodeToUTF(className) + ".class";
    string pjauxFileName =
      fRelativePath + ::UnicodeToUTF(className) + ".pjaux";
    if (fZipFile != 0) {
      ZipDirectory* directory = (ZipDirectory*)fZipFile->central_directory;
      for (int i = 0; i < fZipFile->count;
	   ++i, directory = ZIPDIR_NEXT(directory)) {
	const char* zipFileName = ZIPDIR_FILENAME(directory);
	if (classFileName == zipFileName) {
	  long fileOffset = directory->filestart;
	  long fileLength = directory->size;
	  mapped_file m(fZipFile->fd, fileOffset, fileLength);
	  if (m.is_valid()) {
	    result = CJavaClassFile::ParseFromString(m.get_string());
	    if (result) result->SetDate(fZipDate);
	    break;
	  }
	}
      }
      if (fParameterized && result != 0) {
	directory = (ZipDirectory*)fZipFile->central_directory;
	for (int i = 0; i < fZipFile->count;
	     ++i, directory = ZIPDIR_NEXT(directory)) {
	  const char* zipFileName = ZIPDIR_FILENAME(directory);
	  if (pjauxFileName == zipFileName) {
	    long fileOffset = directory->filestart;
	    long fileLength = directory->size;
	    mapped_file m(fZipFile->fd, fileOffset, fileLength);
	    if (m.is_valid()) {
	      if (! result->ParsePJavaClassInfo(m.get_string())) {
		cerr << "Bad .pjaux file " << pjauxFileName
		     << endl
		     << "(Try re-compiling " << className << ")"
		     << endl;
		result = 0;
	      }
	      break;   
	    }
	  }
	}
      }
    } else {
      string fullPath = fBasePath.GetString() + "/" + classFileName;
      ifstream inFile;
      inFile.open( fullPath.c_str(), ios::in | IOS_BINARY | ios::nocreate);
      if (inFile.is_open()) {
	result = CJavaClassFile::ParseFromFile(inFile);
	inFile.close();
      }
      if (fParameterized && result != 0) {
	string fullPjauxPath = fBasePath.GetString() + "/" + pjauxFileName;
	ifstream inFile;
        inFile.open(fullPjauxPath.c_str(), ios::in | IOS_BINARY | ios::nocreate);
	if (inFile.is_open()) {
	  mapped_file m(inFile.rdbuf()->fd());
	  if (m.is_valid()) {
	    if (! result->ParsePJavaClassInfo(m.get_string())) {
	      cerr << "Bad .pjaux file " << fullPjauxPath
		   << endl
		   << "(Try re-compiling " << className << ")"
		   << endl;
	      result = 0;
	    }
	  }
	}
      }
    }
  }  
  return result;
}

//
//  Method name : HaveFileForClass
//  Description : Returns 'true' iff there seems to be a class file for the 
//     class 'className'.
//
bool 
CJavaDirectory::HaveFileForClass(const unicode_string& className) const
{
  return DateGenerated(className) != (time_t) -1;  
}

//
//  Method name : HaveFileForClass
//  Description : Returns the last modification time for this class file,
//     or -1 if the classfile does not seem to exist.
//
time_t 
CJavaDirectory::DateGenerated(const unicode_string& className) const
{
  if (!IsValid()) return -1;
  string classFileName = fRelativePath + ::UnicodeToUTF(className) + ".class";

  ////
  // Look for this in a zip file.
  ////
  if (fZipFile != 0) {
    ZipDirectory* directory = (ZipDirectory*)fZipFile->central_directory;
    for (int i = 0; i < fZipFile->count;
	 ++i, directory = ZIPDIR_NEXT(directory)) {
      const char* zipFileName = ZIPDIR_FILENAME(directory);
      if (classFileName == zipFileName) {
	struct stat r;
	int result = stat(fBasePath.GetString().c_str(), &r);
	if (result) return -1;
	return r.st_mtime;
      }
    }
  } else {  
    string fullPath = fBasePath.GetString() + "/" + classFileName;
    struct stat r;
    int result = stat(fullPath.c_str(), &r);
    if (result) return -1;
    return r.st_mtime;
  }   
  return -1;
}

//
//  Method name : GetRealDirectory
//  Description : This method will return the actual directory path represented
//    by this path.  This method assumes that this JavaDirectory is not part
//    of a zip file.
//
string
CJavaDirectory::GetRealDirectory() const
{
  assert(!InZipFile());
  if (fRelativePath == "/") return fBasePath.GetString() + "/";
  else return fBasePath.GetString() + "/" + fRelativePath;
}

string
CJavaDirectory::GetZipFileName() const
{
  return fBasePath.GetString();
}

//
//  Method name : GetPackage
//  Description : Returns the portion of this directory that represents the
//    package information.  This is relative to the base directory.
//
string
CJavaDirectory::GetPackage() const
{
  return fRelativePath;
}


//
//  Method name : Convert
//  Description : helper method for converting pathnames
//
string
CJavaDirectory::Convert(string str)
{
  // if necessary, replace all PATH_SEPARATOR with '/', and strip all instances of TOKEN_DELIM
  string newStr="";
  for ( int count=0; count<str.length(); count++ ) {
    switch( str[count] ) {
    case PATH_SEPARATOR:
      if ( USING_CYGNUS ) {
        newStr+='/';
      } else {
        newStr+=PATH_SEPARATOR;
      }
      break;
    case TOKEN_DELIM_CHAR:
      break;
    default:
      newStr += str[count];
    }
  }
  return newStr;
}
